package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;

import meerkat.bulletinboard.workers.multiserver.*;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 05-Dec-15.
 * Thread-based implementation of a Async Bulletin Board Client.
 * Features:
 * 1. Handles tasks concurrently.
 * 2. Retries submitting
 */
public class ThreadedBulletinBoardClient extends SimpleBulletinBoardClient implements AsyncBulletinBoardClient {

    // Executor service for handling jobs
    private final static int JOBS_THREAD_NUM = 5;
    private ExecutorService executorService;

    // Per-server clients
    private List<SingleServerBulletinBoardClient> clients;

    private BatchDigest batchDigest;

    private final static int POST_MESSAGE_RETRY_NUM = 3;
    private final static int READ_MESSAGES_RETRY_NUM = 1;
    private final static int GET_REDUNDANCY_RETRY_NUM = 1;

    private static final int SERVER_THREADPOOL_SIZE = 5;
    private static final long FAIL_DELAY = 5000;
    private static final long SUBSCRIPTION_INTERVAL = 10000;

    private int minAbsoluteRedundancy;

    /**
     * Stores database locations and initializes the web Client
     * Stores the required minimum redundancy.
     * Starts the Thread Pool.
     * @param clientParams contains the required information
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        super.init(clientParams);

        batchDigest = new GenericBatchDigest(digest);

        minAbsoluteRedundancy = (int) (clientParams.getMinRedundancy() * clientParams.getBulletinBoardAddressCount());

        executorService = Executors.newFixedThreadPool(JOBS_THREAD_NUM);

        clients = new ArrayList<>(clientParams.getBulletinBoardAddressCount());
        for (String address : clientParams.getBulletinBoardAddressList()){

            SingleServerBulletinBoardClient client =
                    new SingleServerBulletinBoardClient(SERVER_THREADPOOL_SIZE, FAIL_DELAY, SUBSCRIPTION_INTERVAL);

            client.init(BulletinBoardClientParams.newBuilder()
                            .addBulletinBoardAddress(address)
                            .build());

            clients.add(client);

        }

    }

    /**
     * Post message to all DBs
     * Retry failed DBs
     * @param msg is the message,
     * @return the message ID for later retrieval
     */
    @Override
    public MessageID postMessage(BulletinBoardMessage msg, FutureCallback<Boolean> callback){

        // Create job
        MultiServerPostMessageWorker worker =
                new MultiServerPostMessageWorker(clients, minAbsoluteRedundancy, msg, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

        // Calculate the correct message ID and return it
        batchDigest.reset();
        batchDigest.update(msg.getMsg());
        return batchDigest.digestAsMessageID();

    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {

        // Create job
        MultiServerPostBatchWorker worker =
                new MultiServerPostBatchWorker(clients, minAbsoluteRedundancy, completeBatch, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

        // Calculate the correct message ID and return it
        batchDigest.reset();
        batchDigest.update(completeBatch);
        return batchDigest.digestAsMessageID();

    }

    @Override
    public void beginBatch(BeginBatchMessage beginBatchMessage, FutureCallback<Boolean> callback) {

        // Create job
        MultiServerBeginBatchWorker worker =
                new MultiServerBeginBatchWorker(clients, minAbsoluteRedundancy, beginBatchMessage, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList,
                              int startPosition, FutureCallback<Boolean> callback) {

        BatchDataContainer batchDataContainer = new BatchDataContainer(signerId, batchId, batchDataList, startPosition);

        // Create job
        MultiServerPostBatchDataWorker worker =
                new MultiServerPostBatchDataWorker(clients, minAbsoluteRedundancy, batchDataContainer, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

        postBatchData(signerId, batchId, batchDataList, 0, callback);

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList,
                              int startPosition, FutureCallback<Boolean> callback) {

        postBatchData(signerId.toByteArray(), batchId, batchDataList, startPosition, callback);

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

        postBatchData(signerId, batchId, batchDataList, 0, callback);

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, FutureCallback<Boolean> callback) {

        // Create job
        MultiServerCloseBatchWorker worker =
                new MultiServerCloseBatchWorker(clients, minAbsoluteRedundancy, closeBatchMessage, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    /**
     * Access each database and search for a given message ID
     * Return the number of databases in which the message was found
     * Only try once per DB
     * Ignore communication exceptions in specific databases
     */
    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {

        // Create job
        MultiServerGetRedundancyWorker worker =
                new MultiServerGetRedundancyWorker(clients, minAbsoluteRedundancy, id, GET_REDUNDANCY_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    /**
     * Go through the DBs and try to retrieve messages according to the specified filter
     * If at the operation is successful for some DB: return the results and stop iterating
     * If no operation is successful: return null (NOT blank list)
     */
    @Override
    public void readMessages(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {

        // Create job
        MultiServerReadMessagesWorker worker =
                new MultiServerReadMessagesWorker(clients, minAbsoluteRedundancy, filterList, READ_MESSAGES_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    @Override
    public void readBatch(BatchSpecificationMessage batchSpecificationMessage, FutureCallback<CompleteBatch> callback) {

        // Create job
        MultiServerReadBatchWorker worker =
                new MultiServerReadBatchWorker(clients, minAbsoluteRedundancy, batchSpecificationMessage, READ_MESSAGES_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    @Override
    public void subscribe(MessageFilterList filterList, MessageHandler messageHandler) {
        // TODO: Implement
    }

    @Override
    public void close() {
        super.close();

        try {

            for (SingleServerBulletinBoardClient client : clients){
                client.close();
            }

            executorService.shutdown();
            while (! executorService.isShutdown()) {
                executorService.awaitTermination(10, TimeUnit.SECONDS);
            }
        } catch (InterruptedException e) {
            System.err.println(e.getCause() + " " + e.getMessage());
        }
    }

}
