package meerkat;

import com.google.protobuf.Timestamp;
import meerkat.bulletinboard.AsyncBulletinBoardClient;
import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;
import meerkat.protobuf.VoterRegistry.GroupID;
import meerkat.protobuf.VoterRegistry.VoterID;
import meerkat.protobuf.VoterRegistry.VoterInfo;
import meerkat.protobuf.VoterRegistry.VoterRegistryMessage;
import meerkat.registry.LatestMessagesCallBack;
import meerkat.registry.MessageCollectionUtils;
import meerkat.registry.MessagesCallBack;
import meerkat.registry.RegistryTags;

import java.io.InputStream;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import static meerkat.util.BulletinBoardUtils.signToBulletinBoardMessage;

/**
 * TODO : add ability to use DB of certificates
 */

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class Registry implements VoterRegistry{

    protected Collection<DigitalSignature> signers;
    protected AsyncBulletinBoardClient bulletinBoardClient ;

    @Override
    public void init(Collection<DigitalSignature> signers, AsyncBulletinBoardClient communicator) {
        this.signers = signers;
        this.bulletinBoardClient  = communicator;
    }

    /**
      * Loads all the verification certificate to all the signers
      * @param  certificateStream the certificate for the validation
      */
    public void loadCertificateToSigners(InputStream certificateStream)
    {
        try {
            for (DigitalSignature signer: signers)
            {
                signer.loadVerificationCertificates(certificateStream);
            }
        } catch (CertificateException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void addVoter(VoterInfo voterInfo, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage basicMessage =
                UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + voterInfo.getId().getId())
                        .addTag(RegistryTags.VOTER_ENTRY_TAG)
                        .addTag(voterInfo.getInfo())
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime())
                        .build()).build();

        try {
            bulletinBoardClient.postMessage(signToBulletinBoardMessage(basicMessage, signers),
                    new MessagesCallBack(callback));
        } catch (SignatureException e) {
            callback.handleFailure(e);
        }
    }

    @Override
    public void setVoterGroups(VoterRegistryMessage voterGroup, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterGroup.getVoterID().getId())
                        .addTag(RegistryTags.ADD_TO_GROUP_TAG)
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime())
                        .build());

        for (GroupID groupId : voterGroup.getGroupIDList())
        {
            basicMessage.addTag(RegistryTags.GROUP_ID_TAG + groupId.getId());
        }

        try {
            bulletinBoardClient.postMessage(signToBulletinBoardMessage(basicMessage.build(), signers),
                    new MessagesCallBack(callback));
        } catch (SignatureException e) {
            callback.handleFailure(e);
        }
    }

    @Override
    public void setVoted(VoterID voterId, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterId.getId())
                        .addTag(RegistryTags.VOTE_ACTION_TAG)
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime())
                        .build()).build();

        try {
            bulletinBoardClient.postMessage(signToBulletinBoardMessage(basicMessage, signers),
                    new MessagesCallBack(callback));
        } catch (SignatureException e) {
            callback.handleFailure(e);
        }
    }

    @Override
    public void getGroups(GroupID groupID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>(2) {{
           add(RegistryTags.GROUP_ID_TAG + groupID.getId());
        }};

        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(GroupsActionsTags),
                new LatestMessagesCallBack(callback, signers));
    }

    @Override
    public void getPersonIDDetails(VoterID voterID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>() {{
            add(RegistryTags.ID_TAG + voterID.getId());
            add(RegistryTags.VOTER_ENTRY_TAG);
        }};

        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(GroupsActionsTags),
                new LatestMessagesCallBack(callback, signers));
    }

    @Override
    public boolean hasVoted(VoterID voterId, RegistryCallBack callBack) {
        return false;
    }
}
