package meerkat.registry;

import com.google.common.util.concurrent.FutureCallback;
import meerkat.MessageValidator;
import meerkat.VoterRegistry.RegistryCallBack;
import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.Crypto;
import meerkat.util.TimestampComparator;

import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import static meerkat.util.BulletinBoardUtils.GetListOfTags;

/**
 * TODO : add logging and verification of the messages
 */

/**
 * Created by Vladimir Eliezer Tokarev  on 2/19/2016.
 * Gets latest data from given List<BulletinBoardMessage>
 */
public class LatestMessagesCallBack implements FutureCallback<List<BulletinBoardMessage>>, MessageValidator<List<BulletinBoardMessage>> {
    public RegistryCallBack callback;
    protected Collection<DigitalSignature> validators;

    /**
     * init MessagesCallBack
     * @param callback  voter registry callback object
     * @param validators DigitalSignature object
     */
    public LatestMessagesCallBack(RegistryCallBack callback, Collection<DigitalSignature> validators) {
        this.callback = callback;
        this.validators = validators;
    }

    /**
     * Checks if the given list have tags of type GROUP_ID_TAG or else
     * @param msg List<BulletinBoardAPI.BulletinBoardMessage>
     * @return  true if the messages are with GROUP_ID_TAG tags
     */
    private boolean isAddToGroupsList(List<BulletinBoardMessage> msg) {
        List<String> tags = msg.get(0).getMsg().getTagList();
        for (String tag : tags) {
            if(tag.contains(RegistryTags.GROUP_ID_TAG))
            {
                return true;
            }
        }
        return false;
    }

    /**
     *  Checks if the list of messages is of type GROUP_ID_TAG if it is then calls
     *          handleResult with mapping of the latest groups, else calls to handleResult with
     *          the latest tag from this list (in case of personal data)
     *
     * @param msg List<BulletinBoardAPI.BulletinBoardMessage>
     */
    @Override
    public void onSuccess(List<BulletinBoardMessage> msg) {
        BulletinBoardMessage lastAddedMessage = Collections.max(msg, (first, second) -> {
            TimestampComparator comparator = new TimestampComparator();
            return comparator.compare(first.getMsg().getTimestamp(), second.getMsg().getTimestamp());
        });

        List<BulletinBoardMessage> messages = new ArrayList<BulletinBoardMessage>(){{add(lastAddedMessage);}};

        if (isAddToGroupsList(msg)) {
            callback.handleResult(GetListOfTags(messages, RegistryTags.GROUP_ID_TAG));
        }
        else {
            callback.handleResult(lastAddedMessage);
        }
    }

    /**
     * Calls the callback handleResult method with false because the post method failed
     * @param t the exception data that have been thrown during the failure of the post method
     */
    @Override
    public void onFailure(Throwable t) {
        callback.handleFailure(t);
    }

    /**
     * verify that the message is valid
     * @param message BulletinBoardMessage instance
     * @return true if the message is valid else false
     */
    private boolean verifyMessage(BulletinBoardMessage message)
    {
        try {
            int counter = 0;
            for (Crypto.Signature signature : message.getSigList()) {
                for (DigitalSignature digitalSignature : validators) {
                    if (signature.getSignerId().equals(digitalSignature.getSignerID()))
                    {
                        digitalSignature.initVerify(signature);
                        if(digitalSignature.verify())
                        {
                            counter++;
                        }
                    }
                }
            }
            return counter == validators.size();
        } catch (CertificateException | InvalidKeyException e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * Gets all the valid messages from the list of bulletin board messages
     * @param object object which will be checked
     * @return  List<BulletinBoardMessage>
     * @throws ValidationError
     */
    @Override
    public List<BulletinBoardMessage> validate(List<BulletinBoardMessage> object) throws ValidationError {
        List<BulletinBoardMessage> verifiedMessages = new ArrayList<>(object.size());
        for (BulletinBoardMessage message : object)
        {
            if(verifyMessage(message))
            {
                verifiedMessages.add(message);
            }
        }
        return verifiedMessages;
    }
}
